-module (chat).
-compile (export_all).

% A convenience function to spawn a process for the first
% chat participant.
start_chat () ->
    spawn (?MODULE, first_member, []).

% Function to spawn a chat participant (not the first).
add_member (ExistingMember) ->
    spawn (?MODULE, init_member, [ExistingMember]).

% Setup current process as the only chat member.
first_member () ->
    chat_loop (self ()).

% Initialize a new member by using an existing member.
init_member (ExistingMember) ->
    ExistingMember ! {join, self ()},
    receive {set_neighbour, Neighbour} ->
            chat_loop (Neighbour)
    end.

% Convenience function to send a message.
send (From, To, Msg) ->
    From ! {send_proxy, To, Msg}.

% Convenience function to remove a particular chat member.
remove (ToRemove) ->
    ToRemove ! {remove_proxy}.

% The main chat-loop only remembers one member of the chat.
chat_loop (Neighbour) ->
    receive
        % A new member wants to join, so we use a link-list style
        % to make the new member point-to our current neighbour,
        % and then make the new member our new neighbour.
        {join, NewMember} ->
            NewMember ! {set_neighbour, Neighbour},
            chat_loop (NewMember);
        % If the message is sent to us, print it!
        {send_message, _From, To, Msg} when To == self ()->
            io:format ("> ~s ~n", [Msg]),
            chat_loop (Neighbour);
        % If the message is from us, we report a failure.
        {send_message, From, _To, Msg} when From == self ()->
            io:format ("Message undelivered: ~s~n", [Msg]),
            chat_loop (Neighbour);
        % Otherwise, if it's not for/from us, we send it along the chain.
        {send_message, From, To, Msg} ->
            Neighbour ! {send_message, From, To, Msg},
            chat_loop (Neighbour);
        % Removing a member is done by the member who has the
        % member to remove as a neighbour. We also set out new neighbour
        % which will be the removed-members old neighbour.
        {remove, ToRemove, NewNeighbour} when ToRemove == Neighbour ->
            io:format ("Removing~n"),
            chat_loop (NewNeighbour);
        % If the member to remove we just forward the request along.
        {remove, ToRemove, NewNeighbour} ->
            Neighbour ! {remove, ToRemove, NewNeighbour},
            chat_loop (Neighbour);
        % These are just messages to initiate an external send 
        % and remove request. Effectively, these are an external
        % "message" interface.
        {send_proxy, To, Msg} ->
            Neighbour ! {send_message, self(), To, Msg},
            chat_loop (Neighbour);
        {remove_proxy} ->
            Neighbour ! {remove, self (), Neighbour},
            chat_loop (void);
    end.
